
with Ada.Exceptions;

with Ada.Text_IO;

with Ada.Strings.Fixed;
with Ada.Strings.Maps.Constants;
with Ada.Strings.Bounded;

with Ada.Sequential_IO;

with Ts.Lada;

with Generic_Collection_Support_Logs;
with Generic_Collection_Ordered_Support_Logs;
with Generic_Logs_To_Ram;

package body Logs_To_Ram is
   use Ada;
   use Ts;

   -- -- --
   -- -- --

   --  The following routines do add a separator " | " at the end of the returned string.

   function Str (V : Boolean) return String renames Io_Support.Str;
   function Str (V : Integer) return String renames Io_Support.Str;
   function Str (V : Float) return String renames Io_Support.Str;
   function Str (V : Long_Float) return String renames Io_Support.Str;

   -- -- --

   function Mils (V : Float) return String renames Io_Support.Mils;
   function Mils (V : Long_Float) return String renames Io_Support.Mils;

   -- -- --

   function Put
     (V    : in Float;
      Aft  : in Field := Float'Digits - 1;
      Exp  : in Field := 3)
      return String renames Io_Support.Put;

   function Put
     (V    : in Long_Float;
      Aft  : in Field := Long_Float'Digits - 1;
      Exp  : in Field := 3)
      return String renames Io_Support.Put;

   -- -- --
   -- -- --

   type Log_Item_Type is record
      Ts : Lada.Time_Type; --  Time stamp
      Ms : Bounded_String; --  User message
   end record;

   -- -- --

   type Ordered_Item_Type is record
      Ts : Lada.Time_Type; --  Time stamp
      Tg : Buff_Enum_Type; --  Tag of task
      Nb : Positive; --  Numbering within task
      Ms : Bounded_String; --  User message
   end record;

   function "<" (L, R : Ordered_Item_Type) return Boolean is
      use type Lada.Time_Type;
   begin
      return (L.Ts < R.Ts);
   end "<";

   -- -- --

   package Ordered_Item_To_String is

      function To_String (Ordered_Item : Ordered_Item_Type) return String;

   end Ordered_Item_To_String;

   package body Ordered_Item_To_String is separate;

   -- -- --

   function To_String (Ordered_Item : Ordered_Item_Type) return String renames Ordered_Item_To_String.To_String;

   -- -- --

   function To_Ordered
     (Nb   : Positive;
      T    : Buff_Enum_Type;
      Item : Log_Item_Type)
      return Ordered_Item_Type is
   begin
      return Ordered_Item_Type'(Ts => Item.Ts, Tg => T, Nb => Nb, Ms => Item.Ms);
   end To_Ordered;

   -- -- --

   package Collection_Support is new Generic_Collection_Support_Logs (Log_Item_Type);
   use Collection_Support;

   package Ordered_Collection_Support is new Generic_Collection_Ordered_Support_Logs (Ordered_Item_Type);
   use Ordered_Collection_Support;

   package Log_To_Ram is new Generic_Logs_To_Ram (
      Item_Type                  => Log_Item_Type,
      Collection_Support         => Collection_Support,
      Ordered_Item_Type          => Ordered_Item_Type,
      Ordered_Collection_Support => Ordered_Collection_Support,
      Buff_Enum_Type             => Buff_Enum_Type);

   -- -- --

   Task_Collection_Array : array (Buff_Enum_Type) of Item_Collections_Unbounded.Collection;

   Ordered_Collection : Item_Collections_Ordered_Unbounded.Collection;

   -- -- --

   Ul        : constant Character := '_';
   Base_Name : constant String    := "logging_from_ram";
   Suffix    : constant String    := ".log";

   Output_File : Text_IO.File_Type;

   -- -- --
   -- -- --

   procedure Append_To_Log_Item (B : in Buff_Enum_Type; S : in String) is

   begin
      if not Do_The_Logging then
         return;
      elsif Problems_With_Data_Logging then
         return;
      end if;

      if S = "" then
         return;
      end if;

      Bounded_Ms.Append (Strings_Store_Array (B), S, Strings.Right);

   exception
      when Event : others =>
         Text_IO.Put ("!!!! Logs_To_Ram - Exception with Append_To_Log: ");
         Text_IO.Put_Line (Exceptions.Exception_Information (Event));
         Text_IO.Flush;

         Problems_With_Data_Logging := True;

   end Append_To_Log_Item;

   -- -- --

   procedure Append_To_Log_Item (Pos : in Pos_Nb_Type; S : in String) is

      T : Buff_Enum_Type;
   begin

      T := Buff_Enum_Type'Val (Pos);

      Append_To_Log_Item (T, S);
   end Append_To_Log_Item;

   -- -- --

   procedure Do_Log_Item_To_Ram (B : in Buff_Enum_Type; F : in String := "") is
   begin

      if not Do_The_Logging then
         return;
      elsif Problems_With_Data_Logging then
         return;
      end if;

      if Strings_Store_Array (B) = Null_Bounded_String then
         return;
      end if;

      -- -- --

      declare
         use Ada.Strings;

         S : constant String := To_String (Strings_Store_Array (B));
      begin

         if (F /= "") and then (Fixed.Index (Source => S, Pattern => F, Mapping => Maps.Constants.Lower_Case_Map) = 0) then

            Strings_Store_Array (B) := Null_Bounded_String;

            return;
         end if;

      end; --  declare

      -- -- --

      declare
         Log_Item : Log_Item_Type;

      begin

         Log_Item.Ts := Lada.Clock;
         Log_Item.Ms := Strings_Store_Array (B);

         Log_To_Ram.Save_To_Ram (Task_Collection_Array (B), Log_Item);

         Strings_Store_Array (B) := Null_Bounded_String;

      end; --  declare

   exception
      when Event : others =>
         Text_IO.Put ("!!!! Logs_To_Ram - Exception with Do_Log_All_To_Ram: ");
         Text_IO.Put_Line (Ada.Exceptions.Exception_Information (Event));
         Text_IO.Flush;

         Problems_With_Data_Logging := True;

   end Do_Log_Item_To_Ram;

   -- -- --

   procedure Do_Log_Item_To_Ram (Pos : in Pos_Nb_Type; F : in String := "") is

      B : Buff_Enum_Type;
   begin

      B := Buff_Enum_Type'Val (Pos);

      Do_Log_Item_To_Ram (B => B, F => F);
   end Do_Log_Item_To_Ram;

   -- -- --
   -- -- --

   procedure Do_Log_To_Ram (B : in Buff_Enum_Type; S : in String; F : in String := "") is
      use Ada.Strings;

   begin

      if not Do_The_Logging then
         return;
      elsif Problems_With_Data_Logging then
         return;
      end if;

      if S = "" then
         return;
      end if;

      if (F /= "") and then (Fixed.Index (Source => S, Pattern => F, Mapping => Maps.Constants.Lower_Case_Map) = 0) then
         return;
      end if;

      -- -- --

      declare
         Log_Item : Log_Item_Type;

      begin

         Log_Item.Ts := Lada.Clock;
         Log_Item.Ms := To_Bounded_String (S, Strings.Right);

         Log_To_Ram.Save_To_Ram (Task_Collection_Array (B), Log_Item);

      end; --  declare

   exception
      when Event : others =>
         Text_IO.Put ("!!!! Logs_To_Ram - Exception with Do_Log_To_Ram: ");
         Text_IO.Put_Line (Ada.Exceptions.Exception_Information (Event));
         Text_IO.Flush;

         Problems_With_Data_Logging := True;

   end Do_Log_To_Ram;

   -- -- --

   procedure Do_Log_To_Ram (Pos : in Pos_Nb_Type; S : in String; F : in String := "") is

      T : Buff_Enum_Type;
   begin

      T := Buff_Enum_Type'Val (Pos);

      Do_Log_To_Ram (T, S, F);
   end Do_Log_To_Ram;

   -- -- --
   -- -- --

   procedure Create_Log_File;
   procedure Create_Log_File is
   begin

      declare
         Tstamp : constant String := Ts.Lada.Stamp_Short_Time (1 .. 13);
         --
         --  Without seconds

         Output_Filename : constant String := Base_Name & Ul & Tstamp & Suffix;

      begin
         Text_IO.New_Line;
         Text_IO.Put_Line (">>>> Logs_To_Ram - Create_Log_File ...");
         Text_IO.Flush;

         Text_IO.Create (File => Output_File, Mode => Text_IO.Out_File, Name => Output_Filename);

         Text_IO.Put_Line (">>>> Logs_To_Ram - Create_Log_File.");
         Text_IO.New_Line;
         Text_IO.Flush;

      exception
         when Use_Error =>
            declare
               Output_Filename : constant String := Base_Name & Ul & Tstamp & "_01" & Suffix;
            begin

               Text_IO.Create (File => Output_File, Mode => Text_IO.Out_File, Name => Output_Filename);
            end; --  declare

      end; --  declare

   exception

      when Event : others =>
         Text_IO.Put ("!!!! Logs_To_Ram - Exception with Create_Log_File: ");
         Text_IO.Put_Line (Ada.Exceptions.Exception_Information (Event));
         Text_IO.Flush;

         raise;
   end Create_Log_File;

   -- -- --
   -- -- --

   function Is_With_Logging_Data return Boolean is
   begin

      for Ix in Buff_Enum_Type loop

         if Logs_Buffer.Do_Log_This_Buff_Array (Ix) then

            if not Log_To_Ram.Is_Empty (Task_Collection_Array (Ix)) then

               return True;
            end if;
         end if;

      end loop;

      return False;

   end Is_With_Logging_Data;

   -- -- --
   -- -- --

   procedure Save_Logs_To_File is
   begin

      if not Do_The_Logging then
         return;
      elsif Problems_With_Data_Logging then
         return;
      end if;

      Do_The_Logging := False;

      if not Is_With_Logging_Data then

         Do_The_Logging := True;

         Text_IO.New_Line;
         Text_IO.Put (">>>> Logs_To_Ram - No data was logged.");
         Text_IO.New_Line;
         Text_IO.Flush;

         return;
      end if;

      Create_Log_File;

      for Ix in Buff_Enum_Type loop

         if Logs_Buffer.Do_Log_This_Buff_Array (Ix) then

            Log_To_Ram.Sort_In_Ram (Ix, Task_Collection_Array (Ix), Ordered_Collection);

         end if;

      end loop;

      Log_To_Ram.From_Ram_To_File (Output_File, Ordered_Collection);

      Text_IO.Close (Output_File);

      Log_To_Ram.Clear_Collection (Ordered_Collection);

   exception
      when Event : others =>
         Problems_With_Data_Logging := True;

         Text_IO.Put ("!!!! Logs_To_Ram - Exception with Save_Logs_To_File: ");
         Text_IO.Put_Line (Ada.Exceptions.Exception_Information (Event));
         Text_IO.Flush;

         --  Text_IO.Close (Output_File);

   end Save_Logs_To_File;

   -- -- --
   -- -- --

   function Sep (Nb : Positive := 1) return String is
      use Ada.Strings.Fixed;

      Sep : constant String := "  | ";
   begin
      return Nb * Sep;

   exception
      when others =>
         return Sep;

   end Sep;

-- -- --
-- -- --

begin

   Logs_Buffer.Read_Config_Data (Config_Data => Logs_Buffer.Do_Log_This_Buff_Array);

   for Ix in Buff_Enum_Type loop

      Do_The_Logging := Do_The_Logging or Logs_Buffer.Do_Log_This_Buff_Array (Ix);
   end loop;

exception
   when Event : others =>
      Text_IO.Put ("!!!! Logs_To_Ram - Exception during elaboration: ");
      Text_IO.Put_Line (Ada.Exceptions.Exception_Information (Event));
      Text_IO.Flush;

end Logs_To_Ram;
